@@ -28,17 +28,18 @@
 
 #include <errno.h>
 #include <string.h>
-
-#include <openobex/obex.h>
-#include <openobex/obex_const.h>
+#include <unistd.h>
+#include <inttypes.h>
 
 #include <glib.h>
 
+#include "obexd.h"
 #include "plugin.h"
 #include "obex.h"
 #include "service.h"
 #include "log.h"
-#include "dbus.h"
+#include "manager.h"
+#include "filesystem.h"
 
 #define VCARD_TYPE "text/x-vcard"
 #define VCARD_FILE CONFIGDIR "/vcard.vcf"
@@ -110,34 +111,40 @@ static void opp_progress(struct obex_session *os, void *user_data)
 
 static int opp_chkput(struct obex_session *os, void *user_data)
 {
-	char *folder, *name;
-	char *path;
+	char *folder, *name, *path;
 	int32_t time;
-	int ret;
+	const char *t;
+	int err;
 
 	if (obex_get_size(os) == OBJECT_SIZE_DELETE)
 		return -EINVAL;
 
-	if (obex_get_auto_accept(os)) {
-		folder = g_strdup(obex_get_root_folder(os));
+	t = obex_get_name(os);
+	if (t != NULL && !is_filename(t))
+		return -EBADR;
+
+	if (obex_option_auto_accept()) {
+		folder = g_strdup(obex_option_root_folder());
 		name = g_strdup(obex_get_name(os));
 		goto skip_auth;
 	}
 
 	time = 0;
-	ret = manager_request_authorization(os, time, &folder, &name);
-	if (ret < 0)
+	err = manager_request_authorization(os, time, &folder, &name);
+	if (err < 0)
 		return -EPERM;
 
 	if (folder == NULL)
-		folder = g_strdup(obex_get_root_folder(os));
+		folder = g_strdup(obex_option_root_folder());
 
 	if (name == NULL)
 		name = g_strdup(obex_get_name(os));
 
 skip_auth:
-	if (name == NULL || strlen(name) == 0)
-		return -EBADR;
+	if (name == NULL || strlen(name) == 0) {
+		err = -EBADR;
+		goto failed;
+	}
 
 	if (g_strcmp0(name, obex_get_name(os)) != 0)
 		obex_set_name(os, name);
@@ -146,20 +153,21 @@ skip_auth:
 
 	manager_emit_transfer_started(os);
 
-	ret = obex_put_stream_start(os, path);
+	err = obex_put_stream_start(os, path);
 
 	g_free(path);
+
+failed:
 	g_free(folder);
 	g_free(name);
 
-	return ret;
+	return err;
 }
 
-static int opp_put(struct obex_session *os, obex_object_t *obj,
-						void *user_data)
+static int opp_put(struct obex_session *os, void *user_data)
 {
 	const char *name = obex_get_name(os);
-	const char *folder = obex_get_root_folder(os);
+	const char *folder = obex_option_root_folder();
 
 	if (folder == NULL)
 		return -EPERM;
@@ -170,7 +178,7 @@ static int opp_put(struct obex_session *os, obex_object_t *obj,
 	return 0;
 }
 
-static int opp_get(struct obex_session *os, obex_object_t *obj, void *user_data)
+static int opp_get(struct obex_session *os, void *user_data)
 {
 	const char *type;
 
@@ -182,7 +190,7 @@ static int opp_get(struct obex_session *os, obex_object_t *obj, void *user_data)
 	if (type == NULL)
 		return -EPERM;
 
-	if (g_str_equal(type, VCARD_TYPE)) {
+	if (g_ascii_strcasecmp(type, VCARD_TYPE) == 0) {
 		if (obex_get_stream_start(os, VCARD_FILE) < 0)
 			return -ENOENT;
 
